"""Initailize the wrappers module."""

from credtools.wrappers.abf import run_abf
from credtools.wrappers.abf_cojo import run_abf_cojo
from credtools.wrappers.finemap import run_finemap
from credtools.wrappers.multisusie import run_multisusie
from credtools.wrappers.RSparsePro import run_rsparsepro
from credtools.wrappers.susie import run_susie
from credtools.wrappers.susiex import run_susiex
